//*******************************************************
//	NS73M FM transmitter module + JIHO
//
//	MCU:	ATtiny13V
//			hfuse:FDh	BOD 1.8V
//			lfuse:2Ah	9.6MHz / 8 = 1.2MHz
//
//	FM transmitter module : NS73M (NIIGATA SEIMITSU CO., LTD.)
//
//	Compiler:	AVR-GCC		4.3.4
//				AVR-LIBC	1.6.7
//				binutils	2.20
//
//	2010.12.15	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>

#define		NOP		asm volatile("nop\n"::)
#define		WDR		asm volatile("wdr\n"::)
#define		SLEEP	asm volatile("sleep\n"::)

typedef	unsigned char	uchar;
typedef	unsigned int	uint;
typedef	unsigned long	ulong;


#define		USE_JIHO	0

#if USE_JOHO
#define		EEPROM_CH	(uint8_t *)1
#else
#define		EEPROM_FREQ	(uint16_t *)1
#endif

// |[gݒ(PB)
#define		SDA		4	// PB
#define		SCL		3	// PB
#define		NC1		2
#define		SW		1
#define		LED		0

// TWI(I2C)̃}Nݒ
#define		SCL_H	\
	DDRB &= ~(1<<SCL);	\
	PORTB |= (1<<SCL)

#define		SCL_L	\
	PORTB &= ~(1<<SCL);		\
	DDRB |= (1<<SCL)

#define		SDA_H	\
	DDRB &= ~(1<<SDA);	\
	PORTB |= (1<<SDA)

#define		SDA_L	\
	PORTB &= ~(1<<SDA);		\
	DDRB |= (1<<SDA)

// SCL=HɂȂ܂ő҂
#define		WAIT_SCL_H	while(!(PINB & (1<<SCL)))

void TWI_start(void);
void TWI_send(uchar data);
uchar TWI_recv(uchar ack);
void TWI_stop(void);

void set_data(uchar reg, uchar data);
void wait_timer(int n);

volatile uchar	Timer;

//*******************************************************
//	^C}[Orv`荞
//*******************************************************
ISR(TIM0_COMPA_vect){

	if (Timer)	--Timer;
}


//*******************************************************
//	C[`
//*******************************************************
int main (void) {
	ulong	freq;
	uchar	cex;

	cli();	// 荞݋֎~

	// |[gݒ
	PORTB = (1<<NC1)|(1<<SW);		// vAbv
	DDRB = 0;						// o̓|[g

	// X[vݒ
	//MCUCR = (1<<SE)|(1<<SM1);	// X[v, p[_E[h
	MCUCR = (1<<SE);			// X[v, ACh[h

	// ^C}0ݒ
	TCCR0A = 0;
	TCCR0B = (1<<CS01);		// 1.2MHz/8=150kHz

	// 1bԑ҂
	wait_timer(1000);

	// EEPROMgǂݍ
#if USE_JIHO
	switch(eeprom_read_byte(EEPROM_CH)){
	  case 1:
		freq = 9575;	// TV 1ch
		break;
	  case 3:
		freq = 10775;	// TV 3ch
		break;
	  default:
		freq = 10175;	// TV 2ch
		break;
	}
#else
	freq = (ulong)eeprom_read_word(EEPROM_FREQ);
	if ((freq < 7600)||(freq > 10800)){
		freq = 8960;	// ftHg 89.6MHz
	}
#endif

	// CEXlI
	if (freq < 9000){
		cex = 3;
	} else if (freq < 9600){
		cex = 2;
	} else if (freq < 10200){
		cex = 1;
	} else {
		cex = 0;
	}

	// WX^ݒlvZ
	freq *= 10000;
	freq += 304000;
	freq += 8192/2;
	freq /= 8192;
	
	// NS73M I2CZbg
	TWI_start();
	TWI_send(0xFF);
	TWI_start();
	TWI_stop();

	set_data(0x0E, 0x05);	// Zbg
	set_data(0x01, 0xB4);	// Pilot ON, Forced subcarrier
	set_data(0x02, 0x06);	// ULD ON, 1mW
	set_data(0x03, (uchar)freq);		// g
	set_data(0x04, (uchar)(freq >> 8));	// g
	set_data(0x08, 0x18 + cex);	// Local OSC Band
	set_data(0x00, 0x05);	// input 100mV, Pre-emphasis 50us, Power ON, Mute

	wait_timer(100);

	set_data(0x0E, 0x05);	// Zbg
	set_data(0x06, 0x1E);	// Charge pump

	wait_timer(200);

	set_data(0x00, 0x01);	// Audio input 100mV, Pre-emphasis 50us, Power ON

#if USE_JIHO
	// ȉŃ^C}[ݒύX̂ŁAȍ~ wait_timer() ͎gȂ
	TCCR0A = (1<<COM0A0)|(1<<WGM01);	// CTC, OC0ArvŃgO
	TCCR0B = (1<<CS01);					// vXP[ 1/8
	OCR0A = (uchar)(1200000L/8/(440/2))-1;		// 440Hzo
	TCNT0 = 0;
	TIMSK0 = (1<<OCIE0A);	// rvA荞݋
	
	sei();	// 荞݋

	for (;;){
		char	t;
		uchar	i;

		// XCb`͑҂
		while(PINB & (1<<SW)){
			WDR;
			SLEEP;
		}

		TCNT0 = 0;
		OCR0A = (uchar)(1200000L/8/(440/2))-1;		// 440Hzo

		for (t = 0; t < 3; t++){
			DDRB |= (1<<LED);			// LEDo͋
			Timer = 44*2;		// 0.1b
			while(Timer){
				WDR;
				SLEEP;
			}

			DDRB &= ~(1<<LED);		// LEDo͋֎~
			for (i = 0; i < 9; i++){	// 0.9b
				Timer = 44*2;		// 0.1b
				while(Timer){
					WDR;
					SLEEP;
				}
			}
		}

		OCR0A = (uchar)(1200000L/8/(880/2))-1;		// 880Hzo

		DDRB |= (1<<LED);			// LEDo͋
		for (i = 0; i < 20; i++){
			Timer = 88*2;
			while(Timer){
				WDR;
				SLEEP;
			}
		}
		DDRB &= ~(1<<LED);		// LEDo͋֎~
	}
#else

	// X[vݒ
	MCUCR = (1<<SE)|(1<<SM1);	// X[v, p[_E[h

	DDRB |= (1<<LED);			// LEDo͋
	PORTB |= (1<<LED);

	SLEEP;	// ͐Q

	for (;;){
	}
#endif
}


//*******************************************************
//	1msPʂŎԑ҂
//*******************************************************
void wait_timer(int n){

	while(n > 0){
		WDR;
		// 1msԑ҂
		TCNT0 = 256 - 150;
		TIFR0 |= (1<<TOV0);	// I[o[t[tONA
		while(!(TIFR0 & (1<<TOV0)));
		--n;
	}
}


//************************************************************
//	WX^Ƀf[^Zbg
//************************************************************
void set_data(uchar reg, uchar data){

	TWI_start();
	TWI_send(0xCC);		// AhX + Write
	TWI_send(reg);		// WX^
	TWI_send(data);		// f[^
	TWI_stop();
}


//************************************************************
//	TWI Jn
//************************************************************
void TWI_start(void){

	SDA_H;
	SCL_H;
	SDA_L;
}


//************************************************************
//	TWI 8bit M
//************************************************************
void TWI_send(uchar data){
	char	bit;

	SCL_L;
	for (bit = 0; bit < 8; bit++){
		if (data & 0x80){
			SDA_H;
		} else {
			SDA_L;
		}
		SCL_H;
		WAIT_SCL_H;		// SCL=HɂȂ܂ő҂
		data <<= 1;
		SCL_L;
	}

	SDA_H;
	SCL_L;

	// ACKM e͖
	SCL_H;
	NOP;
	SCL_L;
}


//************************************************************
//	TWI 8bit M
//************************************************************
uchar TWI_recv(uchar nack){
	char	bit;
	uchar	data;

	SDA_H;
	SCL_L;

	data = 0;
	for (bit = 0; bit < 8; bit++){
		SCL_H;
		WAIT_SCL_H;		// SCL=HɂȂ܂ő҂
		data <<= 1;
		if (PINB & (1<<SDA)){
			data |= 1;
		}
		SCL_L;
	}

	// ACK/NACK M
	if (nack){
		SCL_L;
		SCL_H;
		SDA_H;
	} else {
		SDA_L;
		SCL_L;
		SCL_H;
	}

	SCL_L;
	SDA_H;

	return data;
}


//************************************************************
//	TWI ~
//************************************************************
void TWI_stop(void){

	SDA_L;
	NOP;
	SCL_H;
	NOP;
	SDA_H;
}


